package fi.jyu.intellij.comtest.idea.actions

import com.intellij.execution.Executor
import com.intellij.execution.ProgramRunnerUtil
import com.intellij.execution.RunManager
import com.intellij.execution.junit.JUnitConfiguration
import com.intellij.execution.junit.JUnitConfigurationType
import com.intellij.execution.junit.JUnitExternalLibraryDescriptor
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.module.ModuleUtil
import com.intellij.openapi.project.Project
import com.intellij.openapi.roots.ExternalLibraryDescriptor
import com.intellij.openapi.roots.JavaProjectModelModificationService
import com.intellij.openapi.roots.ModuleRootManager
import com.intellij.openapi.vfs.VirtualFileManager
import com.intellij.psi.PsiJavaFile
import com.intellij.psi.PsiManager
import fi.jyu.intellij.comtest.common.util.Notifier
import fi.jyu.intellij.comtest.common.util.runWithRefresh
import fi.jyu.intellij.comtest.idea.ComTestRunner

/**
 * Action to run ComTest on the currently selected/open Java file.
 */
@Suppress("MissingActionUpdateThread", "ComponentNotRegistered")
abstract class ComTestGenerateJavaFileActionBase : AnAction() {
    private var junitDesc: ExternalLibraryDescriptor = JUnitExternalLibraryDescriptor.JUNIT4

    abstract val testExecutor: Executor

    private fun addJUnit(fromFile: PsiJavaFile, proj: Project): Boolean {
        // We add JUnit to the current project
        val fileModule = ModuleUtil.findModuleForFile(fromFile.virtualFile, proj) ?: return Notifier.notifyError(
            proj,
            "No module found for file ${fromFile.name}"
        ).let { false }

        var hasJUnit = false
        ModuleRootManager.getInstance(fileModule).orderEntries()
            .satisfying { it.presentableName == junitDesc.presentableName }
            .forEach {
                hasJUnit = true
                false
            }
        if (!hasJUnit) {
            JavaProjectModelModificationService.getInstance(proj)
                .addDependency(fileModule, JUnitExternalLibraryDescriptor.JUNIT4)
        }
        return true
    }

    override fun actionPerformed(e: AnActionEvent) {
        val proj = e.project ?: return
        val file = e.getData(CommonDataKeys.PSI_FILE) as? PsiJavaFile ?: return
        if (!addJUnit(file, proj)) return

        val generatedFile = runWithRefresh(proj) {
            ComTestRunner.generateForJavaFile(file.virtualFile)
        } ?: return

        // Locate the generated test in the Program Structure Interface and yoink the class name
        // ComTest should only generate one class per file (even if Java would one day allow multiple classes)
        val url = VirtualFileManager.constructUrl("file", generatedFile)
        val generatedFilePsi = VirtualFileManager.getInstance().findFileByUrl(url)
            ?.let { PsiManager.getInstance(proj).findFile(it) as PsiJavaFile? } ?: return

        // Generate the test configuration or update the existing one
        // For now, the qualified name of the class should be unique enough
        val testConfigName = "ComTest: Run for ${file.classes.first().qualifiedName}"
        val runManager = RunManager.getInstance(proj)
        val config = runManager.findConfigurationByName(testConfigName) ?: run {
            val factory = JUnitConfigurationType.getInstance().factory
            val config = runManager.createConfiguration(testConfigName, factory)
            runManager.addConfiguration(config)
            config
        }
        val runConfig = config.configuration as JUnitConfiguration
        runConfig.setMainClass(generatedFilePsi.classes.single())

        ProgramRunnerUtil.executeConfiguration(config, testExecutor)
    }

    override fun update(e: AnActionEvent) {
        e.presentation.isEnabled = e.getData(CommonDataKeys.PSI_FILE) is PsiJavaFile
    }
}